import requests
import json
import os
import datetime
import logging
import typing
import traceback

from typing import Union, List
from requests import Session
from requests.sessions import RequestsCookieJar

from .message import Message
from . import exceptions
from .types.assistant import Assistant
from .types.model import Model
from .types.message import MessageNode, Conversation


conversation = Conversation
model = Model


class ChatBot:
    cookies: dict
    """Cookies for authentication"""

    session: Session
    """HuggingChat session"""

    def __init__(
        self,
        cookies: Union[dict, None, RequestsCookieJar] = None,
        cookie_path: str = "",
        default_llm: Union[int, str] = 0,
        system_prompt: str = "",
    ) -> None:
        """
        Returns a ChatBot object
        default_llm: name or index
        """
        if cookies is None and cookie_path == "":
            raise exceptions.ChatBotInitError(
                "Authentication is required now, but no cookies provided. See tutorial at https://github.com/Soulter/hugging-chat-api"
            )
        elif cookies is not None and cookie_path != "":
            raise exceptions.ChatBotInitError(
                "Both cookies and cookie_path provided")

        if cookies is None and cookie_path != "":
            # read cookies from path
            if not os.path.exists(cookie_path):
                raise exceptions.ChatBotInitError(
                    f"Cookie file {cookie_path} not found. Note: The file must be in JSON format and must contain a list of cookies. See more at https://github.com/Soulter/hugging-chat-api"
                )
            with open(cookie_path, "r", encoding="utf-8") as f:
                cookies = json.load(f)

        # convert cookies to KV format
        if isinstance(cookies, list):
            cookies = {cookie["name"]: cookie["value"] for cookie in cookies}

        self.cookies = cookies

        self.hf_base_url = "https://huggingface.co"
        self.json_header = {"Content-Type": "application/json"}
        self.session = self.get_hc_session()
        self.conversation_list = []
        self.sharing = True
        self.accepted_welcome_modal = (
            False  # It is no longer required to accept the welcome modal
        )

        self.llms = self.get_remote_llms()

        if isinstance(default_llm, str):
            self.active_model = self.get_llm_from_name(default_llm)
            if self.active_model is None:
                raise Exception(
                    f"Given model is not in llms list. LLM list: {[model.id for model in self.llms]}"
                )
        else:
            self.active_model = self.llms[default_llm]

        self.current_conversation = self.new_conversation(
            system_prompt=system_prompt)

    def get_hc_session(self) -> Session:
        session = Session()
        # set cookies
        session.cookies.update(self.cookies)
        session.get(self.hf_base_url + "/chat")
        return session

    def get_headers(self, ref=True, ref_cid: Conversation = None) -> dict:
        _h = {
            "Accept": "*/*",
            "Connection": "keep-alive",
            "Host": "huggingface.co",
            "Origin": "https://huggingface.co",
            "Sec-Fetch-Site": "same-origin",
            "Content-Type": "application/json",
            "Sec-Ch-Ua-Platform": "Windows",
            "Sec-Ch-Ua": 'Chromium";v="116", "Not)A;Brand";v="24", "Microsoft Edge";v="116',
            "Sec-Ch-Ua-Mobile": "?0",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Dest": "empty",
            "Accept-Encoding": "gzip, deflate, br",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36",
        }

        if ref:
            if ref_cid is None:
                ref_cid = self.current_conversation
            _h["Referer"] = f"https://huggingface.co/chat/conversation/{ref_cid}"
        return _h

    def get_cookies(self) -> dict:
        return self.session.cookies.get_dict()

    # NOTE: To create a copy when calling this, call it inside of list().
    #       If not, when updating or altering the values in the variable will
    #       also be applied to this class's variable.
    #       This behavior is with any function returning self.<var_name>. It
    #       acts as a pointer to the data in the object.
    #
    # Returns a pointer to this objects list that contains id of conversations.
    def get_conversation_list(self) -> list:
        return list(self.conversation_list)

    def get_active_llm_index(self) -> int:
        return self.llms.index(self.active_model)

    def accept_ethics_modal(self):
        """
        [Deprecated Method]
        """
        response = self.session.post(
            self.hf_base_url + "/chat/settings",
            headers=self.get_headers(ref=False),
            cookies=self.get_cookies(),
            allow_redirects=True,
            data={
                "ethicsModalAccepted": "true",
                "shareConversationsWithModelAuthors": "true",
                "ethicsModalAcceptedAt": "",
                "activeModel": str(self.active_model),
            },
        )

        if response.status_code != 200:
            raise Exception(
                f"Failed to accept ethics modal with status code: {response.status_code}. {response.content.decode()}"
            )

        return True

    def new_conversation(
        self,
        modelIndex: int = None,
        system_prompt: str = "",
        switch_to: bool = False,
        assistant: Union[str, Assistant] = None,
    ) -> Conversation:
        """
        Create a new conversation. Return a conversation object. 

        modelIndex: int, get it from get_available_llm_models(). If None, use the default model.
        assistant: str or Assistant, the assistant **id** or assistant object. Use search_assistant() to get the assistant object.

        - You should change the conversation by calling change_conversation() after calling this method. Or set param switch_to to True.
        - if you use assistant, the parameter `system_prompt` will be ignored.

        """
        err_count = 0

        if modelIndex is None:
            model = self.active_model
        else:
            if modelIndex < 0 or modelIndex >= len(self.llms):
                raise IndexError("Out of range of llm index")

            model = self.llms[modelIndex]

        # Accept the welcome modal when init.
        # 17/5/2023: This is not required anymore.
        # if not self.accepted_welcome_modal:
        #     self.accept_ethics_modal()

        # Create new conversation and get a conversation id.

        _header = self.get_headers(ref=False)
        _header["Referer"] = "https://huggingface.co/chat"

        request = {
            "model": model.id,
        }

        # get assistant id
        if assistant is not None:
            assistant_id = None
            if isinstance(assistant, str):
                assistant_id = assistant
            elif isinstance(assistant, Assistant):
                assistant_id = assistant.assistant_id
            else:
                raise ValueError(
                    "param assistant must be a string or Assistant object.")
            request["assistantId"] = assistant_id
        else:
            request["preprompt"] = system_prompt if system_prompt != "" else model.preprompt

        while True:
            try:
                resp = self.session.post(
                    self.hf_base_url + "/chat/conversation",
                    json=request,
                    headers=_header,
                    cookies=self.get_cookies(),
                )

                logging.debug(resp.text)
                cid = json.loads(resp.text)["conversationId"]

                c = Conversation(
                    id=cid, system_prompt=system_prompt, model=model)

                self.conversation_list.append(c)
                if switch_to:
                    self.change_conversation(c)

                # we need know the root message id (a.k.a system prompt message id).
                self.get_conversation_info(c)

                return c

            except BaseException as e:
                err_count += 1
                logging.debug(
                    f" Failed to create new conversation ({e}). Retrying... ({err_count})"
                )
                if err_count > 5:
                    raise exceptions.CreateConversationError(
                        f"Failed to create new conversation with status code: {resp.status_code}. Error: {e}. Retries: {err_count}."
                    )
                continue

    def change_conversation(self, conversation_object: Conversation):
        """
        Change the current conversation to another one.
        """

        local_conversation = self.get_conversation_from_id(
            conversation_object.id)

        if local_conversation is None:
            raise exceptions.InvalidConversationIDError(
                "Invalid conversation id, not in conversation list."
            )

        self.get_conversation_info(local_conversation)

        self.current_conversation = local_conversation
        
        return self.current_conversation

    def share_conversation(self, conversation_object: Conversation = None) -> str:
        """
        Return a share link of the conversation.
        """
        if conversation_object is None:
            conversation_object = self.current_conversation

        headers = self.get_headers()

        r = self.session.post(
            f"{self.hf_base_url}/chat/conversation/{conversation_object}/share",
            headers=headers,
            cookies=self.get_cookies(),
        )

        if r.status_code != 200:
            raise Exception(
                f"Failed to share conversation with status code: {r.status_code}"
            )

        response = r.json()
        if "url" in response:
            return response["url"]

        raise Exception(f"Unknown server response: {response}")

    def delete_all_conversations(self) -> None:
        """
        Deletes ALL conversations on the HuggingFace account
        """

        settings = {"": ("", "")}

        r = self.session.post(
            f"{self.hf_base_url}/chat/conversations?/delete",
            headers={"Referer": "https://huggingface.co/chat", "Origin": "https://huggingface.co", "Content-Type": "multipart/form-data; boundary=----WebKitFormBoundarywrIEW0Ame78HYisT"},
            cookies=self.get_cookies(),
            allow_redirects=True,
            files=settings,
        )

        if r.status_code != 200:
            raise exceptions.DeleteConversationError(
                f"Failed to delete ALL conversations with status code: {r.status_code}"
            )

        self.conversation_list = []
        self.current_conversation = None

    def delete_conversation(self, conversation_object: Conversation = None) -> None:
        """
        Delete a HuggingChat conversation by conversation.
        """

        if conversation_object is None:
            conversation_object = self.current_conversation

        headers = self.get_headers()

        r = self.session.delete(
            f"{self.hf_base_url}/chat/conversation/{conversation_object}",
            headers=headers,
            cookies=self.get_cookies(),
        )

        if r.status_code != 200:
            raise exceptions.DeleteConversationError(
                f"Failed to delete conversation with status code: {r.status_code}"
            )
        else:
            self.conversation_list.pop(
                self.get_conversation_from_id(
                    conversation_object.id, return_index=True)
            )

            if conversation_object is self.current_conversation:
                self.current_conversation = None

    def get_available_llm_models(self) -> list:
        """
        Get all available models that are available in huggingface.co/chat.
        """
        return self.llms

    def set_share_conversations(self, val: bool = True):
        """
        Sets the "Share Conversation with Model Authors setting" to the given val variable
        """
        settings = {"shareConversationsWithModelAuthors": (
            "", "on" if val else "")}

        r = self.session.post(
            self.hf_base_url + "/chat/settings",
            headers={"Referer": "https://huggingface.co/chat", "Origin":"https://huggingface.co"},
            cookies=self.get_cookies(),
            allow_redirects=True,
            files=settings,
        )

        if r.status_code != 200:
            raise Exception(
                f"Failed to set share conversation with status code: {r.status_code}"
            )

    def switch_llm(self, index: int) -> bool:
        """
        Attempts to change current conversation's Large Language Model.
        Requires an index to indicate the model you want to switch.
        See self.llms for available models.

        Note: 1. The effect of switch is limited to the current conversation,
        You can manually switch the llm when you start a new conversation.

        2. Only works *after creating a new conversation.*
        :)
        """
        # TODO: I will work on making it have a model for each conversation that is changeable. - @Zekaroni

        if index < len(self.llms) and index >= 0:
            self.active_model = self.llms[index]
            return True
        else:
            raise IndexError("Out of range of llm index")

    def get_llm_from_name(self, name: str) -> Union[Model, None]:
        for model in self.llms:
            if model.name == name:
                return model

    # Gives information such as name, websiteUrl, description, displayName, parameters, etc.
    # We can use it in the future if we need to get information about models
    def get_remote_llms(self) -> list:
        """
        Fetches all possible LLMs that could be used. Returns the LLMs in a list
        """

        r = self.session.post(
            self.hf_base_url + "/chat/__data.json",
            headers=self.get_headers(ref=False),
            cookies=self.get_cookies(),
        )

        if r.status_code != 200:
            raise Exception(
                f"Failed to get remote LLMs with status code: {r.status_code}"
            )

        data = r.json()["nodes"][0]["data"]
        modelsIndices = data[data[0]["models"]]
        model_list = []

        def return_data_from_index(
            index): return None if index == -1 else data[index]

        for modelIndex in modelsIndices:
            model_data = data[modelIndex]

            # Model is unlisted, skip it
            if data[model_data["unlisted"]]:
                continue

            m = Model(
                id=return_data_from_index(model_data["id"]),
                name=return_data_from_index(model_data["name"]),
                displayName=return_data_from_index(model_data["displayName"]),
                preprompt=return_data_from_index(model_data["preprompt"]),
                # promptExamples = return_data_from_index(model_data["promptExamples"]),
                websiteUrl=return_data_from_index(model_data["websiteUrl"]),
                description=return_data_from_index(model_data["description"]),
                datasetName=return_data_from_index(model_data["datasetName"]),
                datasetUrl=return_data_from_index(model_data["datasetUrl"]),
                modelUrl=return_data_from_index(model_data["modelUrl"]),
                # parameters = return_data_from_index(model_data["parameters"]),
            )

            prompt_list = return_data_from_index(model_data["promptExamples"])
            if prompt_list is not None:
                _promptExamples = [
                    return_data_from_index(index) for index in prompt_list
                ]
                m.promptExamples = [
                    {"title": data[prompt["title"]],
                        "prompt": data[prompt["prompt"]]}
                    for prompt in _promptExamples
                ]

            indices_parameters_dict = return_data_from_index(
                model_data["parameters"])
            out_parameters_dict = {}
            for key, value in indices_parameters_dict.items():
                if value == -1:
                    out_parameters_dict[key] = None
                    continue

                if isinstance(type(data[value]), list):
                    out_parameters_dict[key] = [data[index]
                                                for index in data[value]]
                    continue

                out_parameters_dict[key] = data[value]

            m.parameters = out_parameters_dict

            model_list.append(m)

        return model_list

    def get_remote_conversations(self, replace_conversation_list=True):
        """
        Returns all the remote conversations for the active account. Returns the conversations in a list.
        """

        r = self.session.post(
            self.hf_base_url + "/chat/__data.json",
            headers=self.get_headers(ref=False),
            cookies=self.get_cookies(),
        )

        if r.status_code != 200:
            raise Exception(
                f"Failed to get remote conversations with status code: {r.status_code}"
            )

        data = r.json()["nodes"][0]["data"]
        conversationIndices = data[data[0]["conversations"]]
        conversations = []

        for index in conversationIndices:
            conversation_data = data[index]
            c = Conversation(
                id=data[conversation_data["id"]],
                title=data[conversation_data["title"]],
                model=data[conversation_data["model"]],
            )

            conversations.append(c)

        if replace_conversation_list:
            self.conversation_list = conversations

        return conversations

    def get_conversation_info(self, conversation: Union[Conversation, str] = None) -> Conversation:
        """
        Fetches information related to the specified conversation. Returns the conversation object.
        conversation: Conversation object that has the conversation id Or None to use the current conversation.
        """

        if conversation is None:
            conversation = self.current_conversation

        if isinstance(conversation, str):
            conversation = Conversation(id=conversation)

        r = self.session.get(
            self.hf_base_url +
            f"/chat/conversation/{conversation.id}/__data.json?x-sveltekit-invalidated=01",
            headers=self.get_headers(ref=False),
            cookies=self.get_cookies(),
        )

        if r.status_code != 200:
            raise Exception(
                f"Failed to get conversation info with status code: {r.status_code}"
            )
            
        # you'll never understand the following codes until you try to debug huggingchat in person.
        data = r.json()["nodes"][1]["data"]

        conversation.model = data[data[0]["model"]]
        conversation.system_prompt = data[data[0]["preprompt"]]
        conversation.title = data[data[0]["title"]]

        messages: list = data[data[0]["messages"]]
        conversation.history = []

        # parse all message nodes (history) in the conversation
        for index in messages:  # node's index
            _node_meta = data[index]
            # created_at and updated_at may not in the metadata of a message node.
            created_at = datetime.datetime.strptime(
                    data[_node_meta["createdAt"]][1], "%Y-%m-%dT%H:%M:%S.%fZ").timestamp() if 'createdAt' in _node_meta else None
            updated_at = datetime.datetime.strptime(
                    data[_node_meta["updatedAt"]][1], "%Y-%m-%dT%H:%M:%S.%fZ").timestamp() if 'updatedAt' in _node_meta else None
            ancestors = []
            children = []
            for a_idx in data[_node_meta["ancestors"]]:
                ancestors.append(data[a_idx])
            for c_idx in data[_node_meta["children"]]:
                children.append(data[c_idx])
            conversation.history.append(MessageNode(
                id=data[_node_meta["id"]],
                role=data[_node_meta["from"]],
                content=data[_node_meta["content"]],
                ancestors=ancestors,
                children=children,
                created_at=created_at,
                updated_at=updated_at,
            ))

        logging.debug(
            f"conversation {conversation.id} history: {conversation.history}")

        return conversation

    def get_conversation_from_id(self, conversation_id: str, return_index=False) -> Conversation:
        """
        Returns a conversation object that is already in the conversation list.
        """

        for i, conversation in enumerate(self.conversation_list):
            if conversation.id == conversation_id:
                if return_index:
                    return i
                return conversation

    def _parse_assistants(self, nodes_data: list) -> List[Assistant]:
        '''
        parse the assistants data from the response.
        '''
        index = nodes_data[1]
        ret = []
        for i in index:
            attribute_map: dict = nodes_data[i]
            assistant_id = nodes_data[attribute_map['_id']]
            author = nodes_data[attribute_map['createdByName']]
            name = nodes_data[attribute_map['name']].strip()
            model_name = nodes_data[attribute_map['modelId']]
            pre_prompt = nodes_data[attribute_map['preprompt']]
            description = nodes_data[attribute_map['description']]
            ret.append(Assistant(
                assistant_id,
                author,
                name,
                model_name,
                pre_prompt,
                description
            ))
        return ret

    def get_assistant_list_by_page(self, page: int) -> List[Assistant]:
        '''
        get assistant list by page number.
        if page < 0 or page > max_page then return `None`.
        '''
        url_cache = f"https://api.soulter.top/hugchat/assistants/__data.json?p={page}"
        url = f"https://huggingface.co/chat/assistants/__data.json?p={page}&x-sveltekit-invalidated=01"
        try:
            res = requests.get(url_cache, timeout=5)
        except BaseException:
            res = self.session.get(url, timeout=10)
        res = res.json()
        if res['nodes'][1]['type'] == 'error':
            return None
        # here we parse the result
        return self._parse_assistants(res['nodes'][1]['data'])

    def search_assistant(self, assistant_name: str = None, assistant_id: str = None) -> Assistant:
        '''
        - If you created an assistant by your own, you should pass the assistant_id here but not the assistant_name. You can pass your assistant_id into the new_conversation() directly.
        - Search an available assistant by assistant name or assistant id.
        - Will search on api.soulter.top/hugchat because offifial api doesn't support search.
        - Return the `Assistant` object if found, return None if not found.
        '''
        if not assistant_name and not assistant_id:
            raise ValueError(
                "assistant_name and assistant_id can not be both None.")
        if assistant_name:
            url = f"https://api.soulter.top/hugchat/assistant?name={assistant_name}"
        else:
            url = f"https://api.soulter.top/hugchat/assistant?id={assistant_id}"
        res = requests.get(url, timeout=10)
        if res.status_code != 200:
            raise Exception(
                f"Failed to search assistant with status code: {res.status_code}, please commit an issue to https://github.com/Soulter/hugging-chat-api/issues")
        res = res.json()
        if not res['data']:
            # empty dict
            return None
        if res['code'] != 0:
            raise Exception(
                f"Failed to search assistant with server's error: {res['message']}, please commit an issue to https://github.com/Soulter/hugging-chat-api/issues")
        return Assistant(**res['data'])

    def _stream_query(
        self,
        text: str,
        web_search: bool = False,
        is_retry: bool = False,
        retry_count: int = 5,
        conversation: Conversation = None,
        message_id: str = None,
    ) -> typing.Generator[dict, None, None]:
        if conversation is None:
            conversation = self.current_conversation

        if retry_count <= 0:
            raise Exception(
                "the parameter retry_count must be greater than 0.")
        if len(conversation.history) == 0:
            raise Exception(
                "conversation history is empty, but we need the root message id of this conversation to continue.")
            
        if not message_id:
            # get last message id
            message_id = conversation.history[-1].id
            
        logging.debug(f'message_id: {message_id}')

        req_json = {
            "id": message_id,
            "inputs": text,
            "is_continue": False,
            "is_retry": is_retry,
            "web_search": web_search,
            "tools": {}
        }
        headers = {
            'authority': 'huggingface.co',
            'accept': '*/*',
            'accept-language': 'zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7,en-GB;q=0.6',
            'origin': 'https://huggingface.co',
            'sec-ch-ua': '"Microsoft Edge";v="119", "Chromium";v="119", "Not?A_Brand";v="24"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"macOS"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36 Edg/119.0.0.0',
        }
        final_answer = {}

        break_flag = False

        while retry_count > 0:
            resp = self.session.post(
                self.hf_base_url + f"/chat/conversation/{conversation}",
                files={ "data": (None, json.dumps(req_json)) },
                stream=True,
                headers=headers,
                cookies=self.session.cookies.get_dict(),
            )
            resp.encoding = 'utf-8'

            if resp.status_code != 200:
                
                retry_count -= 1
                if retry_count <= 0:
                    raise exceptions.ChatError(
                        f"Failed to chat. ({resp.status_code})")

            try:
                for line in resp.iter_lines(decode_unicode=True):
                    if not line:
                        continue
                    res = line
                    obj = json.loads(res)
                    if obj.__contains__("type"):
                        _type = obj["type"]

                        if _type == "finalAnswer":
                            final_answer = obj
                            break_flag = True
                            break
                    else:
                        logging.error(f"No `type` found in response: {obj}")
                    yield obj
            except requests.exceptions.ChunkedEncodingError:
                pass
            except BaseException as e:
                traceback.print_exc()
                if "Model is overloaded" in str(e):
                    raise exceptions.ModelOverloadedError(
                        "Model is overloaded, please try again later or switch to another model."
                    )
                logging.debug(resp.headers)
                if "Conversation not found" in str(res):
                    raise exceptions.InvalidConversationIDError("Conversation id invalid")
                raise exceptions.ChatError(f"Failed to parse response: {res}")
            if break_flag:
                break

        # update the history of current conversation
        self.get_conversation_info(conversation)
        yield final_answer

    def query(self) -> Message:
        """
        **Deprecated**
        Please use `chat()`. The function will raise an error immediately.
        """
        raise Exception("The function is deprecated. Please use `chat()`")
    
    def get_message_node(self, conversation: Conversation, message_id: str):
        for node in conversation.history:
            if node.id == message_id:
                return node
        raise Exception(f"no node found which id is {message_id}")

    def chat(
        self,
        text: str,
        web_search: bool = False,
        # For stream mode, yield all responses from the server.
        _stream_yield_all: bool = False,
        retry_count: int = 5,
        conversation: Conversation = None,
        edit_user_node: MessageNode = None,
        *args,
        **kvargs,
    ) -> Message:
        """
        - Send a message to the current conversation. Return a Message object.

        - You can turn on the web search by set the parameter `web_search` to True.

        - Stream is now the default mode, you can call Message.wait_until_done() to get the result text.
        
        - `Edit history`: pass `edit_user_node`. The history can be retrieved from `conversation.history`. 

        - About class `Message`:
            - `wait_until_done()`: Block until the response done processing or an error raised.
            - `__iter__()`: For loop call this Generator and get response.
            - `get_search_sources()`: The web search results. It is a list of WebSearchSource objects.

        - For more detail please see Message documentation(Message.__doc__)
        """
        if conversation is None:
            conversation = self.current_conversation
        
        if not text:
            raise Exception("don't support an empty string(I'm sure LLM cannot understand it)")
        if edit_user_node and edit_user_node.role != 'user':
            raise Exception("you must pass a user's message node to edit message")

        is_retry = True if edit_user_node else False
        edit_message_id = edit_user_node.id if is_retry else None

        msg = Message(
            g=self._stream_query(
                text=text,
                web_search=web_search,
                retry_count=retry_count,
                conversation=conversation,
                is_retry=is_retry,
                message_id=edit_message_id
            ),
            _stream_yield_all=_stream_yield_all,
            web_search=web_search,
            conversation=conversation
        )
        return msg


if __name__ == "__main__":
    bot = ChatBot()
    message_content = bot.chat("Hello")
    print(message_content)
    sharelink = bot.share_conversation()
    print(sharelink)



from dataclasses import dataclass
from typing import List
from .model import Model

@dataclass
class MessageNode:
    '''
    huggingchat message node, currently only maintain id, role, date and content.
    '''
    id: str
    role: str  # "user", "system", or "assistant"
    content: str
    ancestors: List[str]
    children: List[str]
    created_at: float  # timestamp
    updated_at: float  # timestamp

    def __str__(self) -> str:
        return f"MessageNode(id={self.id}, role={self.role}, content={self.content}, created_at={self.created_at}, updated_at={self.updated_at})"


class Conversation:
    def __init__(
        self,
        id: str = None,
        title: str = None,
        model: Model = None,
        system_prompt: str = None,
        history: list = []
    ):
        """
        Returns a conversation object
        """

        self.id: str = id
        self.title: str = title
        self.model = model
        self.system_prompt: str = system_prompt
        self.history: List[MessageNode] = history

    def __str__(self) -> str:
        return self.id



from dataclasses import dataclass


@dataclass
class Tool:
    '''
    Class used to represent tools used by the model
    '''

    uuid: str
    result: str

    def __str__(self) -> str:
        return f"Tool(uuid={self.uuid}, result={self.result})"



import requests
import os
import json
import logging
import re


class Login:

    def __init__(self, email: str, passwd: str = "") -> None:
        self.DEFAULT_PATH_DIR = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "usercookies")
        self.DEFAULT_COOKIE_PATH = self.DEFAULT_PATH_DIR + \
            os.path.join(f"{email}.json")

        self.email: str = email
        self.passwd: str = passwd
        self.headers = {
            "Referer": "https://huggingface.co/",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Safari/537.36 Edg/112.0.1722.64",
        }
        self.cookies = requests.sessions.RequestsCookieJar()

    def login(self, cookie_dir_path: str = None, save_cookies: bool = False) -> requests.sessions.RequestsCookieJar:
        '''
        Login to huggingface.co with given email and password.
        - If cookie_dir_path is given, load cookies from the path. If path is not exist, raise an exception.
        - If save_cookies is True, save cookies to the givev path (if cookie_dir_path is not given, save to default path `./usercookies`).
        - Return cookies if login success, otherwise raise an exception.
        '''

        if not cookie_dir_path:
            cookie_dir_path = self.DEFAULT_PATH_DIR

        # validate cookies content before use
        if os.path.exists(cookie_dir_path) and os.path.exists(self._get_cookie_path(cookie_dir_path)):
            with open(self._get_cookie_path(cookie_dir_path), 'r+') as f_cookies:
                cookies = json.load(f_cookies)
                try:
                    list(cookies.keys()).index('token')
                    list(cookies.keys()).index('hf-chat')
                    return self.load_cookies(cookie_dir_path)
                except Exception as e:
                    print('error during validating cookies')

        self._sign_in_with_email()
        location = self._get_auth_url()
        if self._grant_auth(location):
            if save_cookies:
                self.save_cookies(cookie_dir_path)
            return self.cookies
        else:
            raise Exception(
                f"Grant auth fatal, please check your email or password\ncookies gained: \n{self.cookies}")

    def save_cookies(self, cookie_dir_path: str = './usercookies') -> str:
        '''
        cookies will be saved into: cookie_dir_path/<email>.json
        '''
        return self.saveCookiesToDir(cookie_dir_path)

    def saveCookiesToDir(self, cookie_dir_path: str = './usercookies') -> str:
        """
        alias of save_cookies
        """
        cookie_dir_path = self.DEFAULT_PATH_DIR if not cookie_dir_path else cookie_dir_path
        if not cookie_dir_path.endswith("/"):
            cookie_dir_path += "/"
        cookie_path = cookie_dir_path + f"{self.email}.json"
        if not os.path.exists(cookie_dir_path):
            logging.info("Cookie directory not exist, creating...")
            os.makedirs(cookie_dir_path)
        logging.info(f"Cookie store path: {cookie_path}")

        with open(cookie_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(self.cookies.get_dict()))
        return cookie_path

    def load_cookies(self, cookie_dir_path: str = './usercookies') -> requests.sessions.RequestsCookieJar:
        '''
        cookies will be loaded from: cookie_dir_path/<email>.json
        '''
        return self.loadCookiesFromDir(cookie_dir_path)

    def loadCookiesFromDir(self, cookie_dir_path: str = './usercookies') -> requests.sessions.RequestsCookieJar:
        """
        alias of load_cookies
        """
        cookie_dir_path = self.DEFAULT_PATH_DIR if not cookie_dir_path else cookie_dir_path
        cookie_path = self._get_cookie_path(cookie_dir_path)
        if not cookie_path:
            raise Exception(f"Cookie not found. please check the path given: {cookie_dir_path}.\n" +
                            f"Cookie file must be named like this: 'your_email'+'.json': '{self.email}.json'")

        with open(cookie_path, "r", encoding="utf-8") as f:
            try:
                js = json.loads(f.read())
                for i in js.keys():
                    self.cookies.set(i, js[i])
                    logging.info(f"{i} loaded")
                return self.cookies
            except:
                raise Exception(
                    "load cookies from files fatal. Please check the format")

    def _request_get(self, url: str, params=None, allow_redirects=True) -> requests.Response:
        res = requests.get(
            url,
            params=params,
            headers=self.headers,
            cookies=self.cookies,
            allow_redirects=allow_redirects,
        )
        self._refresh_cookies(res.cookies)
        return res

    def _request_post(self, url: str, headers=None, params=None, data=None, stream=False,
                      allow_redirects=True) -> requests.Response:
        res = requests.post(
            url,
            stream=stream,
            params=params,
            data=data,
            headers=self.headers if headers is None else headers,
            cookies=self.cookies,
            allow_redirects=allow_redirects
        )
        self._refresh_cookies(res.cookies)
        return res

    def _refresh_cookies(self, cookies: requests.sessions.RequestsCookieJar):
        dic = cookies.get_dict()
        for i in dic:
            self.cookies.set(i, dic[i])

    def _sign_in_with_email(self):
        """
        Login through your email and password.
        PS: I found that it doesn't have any type of encryption till now,
        which could expose your password to the internet.
        """
        url = "https://huggingface.co/login"
        data = {
            "username": self.email,
            "password": self.passwd,
        }
        res = self._request_post(url=url, data=data, allow_redirects=False)
        if res.status_code == 400:
            raise Exception("wrong username or password")

    def _get_auth_url(self):
        url = "https://huggingface.co/chat/login"
        headers = {
            "Referer": "https://huggingface.co/chat/login",
            "User-Agent": self.headers["User-Agent"],
            "Content-Type": "application/x-www-form-urlencoded",
            "Origin": "https://huggingface.co/chat"
        }
        res = self._request_post(url, headers=headers, allow_redirects=False)
        if res.status_code == 200:
            # location = res.headers.get("Location", None)
            location = res.json()["location"]
            if location:
                return location
            else:
                raise Exception(
                    "No authorize url found, please check your email or password.")
        elif res.status_code == 303:
            location = res.headers.get("Location")
            if location:
                return location
            else:
                raise Exception(
                    "No authorize url found, please check your email or password.")
        else:
            raise Exception("Something went wrong!")

    def _grant_auth(self, url: str) -> int:
        res = self._request_get(url, allow_redirects=False)
        if res.headers.__contains__("location"):
            location = res.headers["location"]
            res = self._request_get(location, allow_redirects=False)
            if res.cookies.__contains__("hf-chat"):
                return 1
        # raise Exception("grantAuth fatal")
        if res.status_code != 200:
            raise Exception("grant auth fatal!")
        csrf = re.findall(
            '/oauth/authorize.*?name="csrf" value="(.*?)"', res.text)
        if len(csrf) == 0:
            raise Exception("No csrf found!")
        data = {
            "csrf": csrf[0]
        }

        res = self._request_post(url, data=data, allow_redirects=False)
        if res.status_code != 303:
            raise Exception(f"get hf-chat cookies fatal! - {res.status_code}")
        else:
            location = res.headers.get("Location")
        res = self._request_get(location, allow_redirects=False)
        if res.status_code != 302:
            raise Exception(f"get hf-chat cookie fatal! - {res.status_code}")
        else:
            return 1

    def _get_cookie_path(self, cookie_dir_path) -> str:
        if not cookie_dir_path.endswith("/"):
            cookie_dir_path += "/"
        if not os.path.exists(cookie_dir_path):
            return ""
        files = os.listdir(cookie_dir_path)
        for i in files:
            if i == f"{self.email}.json":
                return cookie_dir_path + i
        return ""



"""Entry for the hugchat command line interface.

simply type
```bash
python -m hugchat.cli
```
to start cli.
"""

from .hugchat import ChatBot
from .login import Login
import getpass
import argparse
import os
import traceback

ENVIRONMENT_EMAIL = os.getenv("EMAIL")
ENVIRONMENT_PASSWORD = os.getenv("PASSWD")
stream_output = False
is_web_search = False
web_search_hint = False
continued_conv = False

EXIT_COMMANDS = [
    "/exit",
    "/quit",
    "/stop",
    "/close"
]

# COOKIE_PATH_DIR = os.path.abspath(os.path.dirname(__file__)) + "/usercookies"


def handle_command(chatbot: ChatBot, userInput: str) -> None:
    global stream_output, is_web_search, web_search_hint , continued_conv

    arguments = userInput.lower().split(" ")
    command = arguments[0][1:] # Remove the '/' at the start of the input
    arguments = arguments[1:]

    if command == "help" or command == "commands":
        print("""
/new : Create and switch to a new conversation.
/ids : Shows a list of all ID numbers and ID strings in *current session*.
/switch : Shows a list of all conversations' info in *current session*. Then you can choose one to switch to.
/switch all : Shows a list of all conversations' info in *your account*. Then you can choose one to switch to. (not recommended if your account has a lot of conversations)
/del <index> : Deletes the conversation linked with the index passed. Will not delete active session.
/delete-all : Deletes all the conversations for the logged in user.
/clear : Clear the terminal.
/llm : Get available models you can switch to.
/llm <index> : Switches model to given model index based on /llm.
/share : Toggles settings for sharing data with model author. On by default.
/exit : Closes CLI environment.
/stream : Toggles streaming the response.
/web : Toggles web search.
/web-hint : Toggles display web search hint.
        """)

    elif command == "new":
        new_conversation = chatbot.new_conversation(switch_to=True)
        conversation_index = len(chatbot.get_conversation_list())
        print(f"# Created and switched to a new conversation\n# New conversation ID: {new_conversation.id}\n# New conversation index: {conversation_index}")

    elif command == "ids":
        print(f"# Conversations: ")
        for i, conversation in enumerate(chatbot.get_conversation_list()):
            print(f"# {i+1}.) {conversation.id}{' (acitve)' if chatbot.get_conversation_info().id == conversation.id else ''}")

    elif command == "switch":
        try:
            if arguments and arguments[0] == "all":
                id = chatbot.get_remote_conversations(replace_conversation_list=True)
            elif not arguments:
                id = chatbot.get_conversation_list()
            else:
                print("# Invalid argument(s).")
                return
            conversation_dict = {i+1: id_string for i, id_string in enumerate(id)}

            for i, id_string in conversation_dict.items():
                info = chatbot.get_conversation_info(id_string)
                print(f"# {i}: ID: {info.id}\n# Title: {info.title[:43]}...\n# Model: {info.model}.\n# System Prompt: {info.system_prompt}\n# --------------------------------------------------------")

            index_value = int(input("# Choose conversation ID(input the index): "))
            target_id = conversation_dict.get(index_value)

            if target_id:
                chatbot.change_conversation(target_id)
                print(f"# Switched to conversation with ID: {target_id}\n")
            else:
                print("# Invalid conversation ID")
        except Exception as e:
            if isinstance(e, ValueError):
                print(f"# Invlid argument.")
            else:
                print(f"# Error: {e}")

    elif command == "del" or command == "delete":
        try:
            if not arguments:
                print("# No argument for index provided.")
                return

            conversation_index = int(arguments[0])
            if conversation_index < 1:
                raise IndexError()
            selected_conversation = chatbot.get_conversation_list()[conversation_index-1].id
            current_conversation_id = chatbot.get_conversation_info().id

            if selected_conversation == current_conversation_id:
                print("# Cannot delete active chat.")
                return
            
            to_delete_conversation = chatbot.get_conversation_from_id(selected_conversation)
        except Exception as e:
            if isinstance(e, ValueError):
                print("# Invalid argument for index.")
            else:
                print("# Invalid index.")
            return
        
        if not to_delete_conversation is None:
            chatbot.delete_conversation(to_delete_conversation)
            print("# Deleted conversation successfully.")
            return
        print("# Error")

    elif command == "delete-all" or command == "deleteall":
        if input("# WARNING: This will delete all conversations linked with this account. Continue? (y/n) : ").lower() == 'y':
            chatbot.delete_all_conversations()
            print("# Deleted all conversations successfully")

            new_conversation = chatbot.new_conversation(switch_to=True)
            print(f"# Created and switched to a new conversation\n# New conversation ID: {new_conversation.id}")

    elif command == "clear" or command == "cls":
        os.system('cls' if os.name == 'nt' else 'clear')

    elif command == "llm":
        if len(arguments) == 0:
            print(f"# Available Models: ")
            for i, model in enumerate(chatbot.get_available_llm_models()):
                print(f"# {i+1}.) {model.id}")
            return

        try:
            chatbot.switch_llm(int(arguments[0])-1)
        except Exception as e:
            if isinstance(e, ValueError):
                print("# Not a valid argument")
            elif isinstance(e, IndexError):
                print("# Invalid LLM index")
            else:
                print(e)
            return

        print(f"# Switched to LLM {chatbot.active_model.id}\n# Please note that you have to create a new conversation for this to take effect")

    elif command == "share":
        if arguments:
            chatbot.sharing = arguments[0] == "on"
        else:
            chatbot.sharing = not chatbot.sharing
        try:
            chatbot.set_share_conversations(chatbot.sharing)
            print(f"# {'Now sharing conversations with model author' if chatbot.sharing else 'No longer sharing conversations with model author'}")
        except Exception as e:
            print(f"# Error: {e}\n")

    elif command == "stream" or command == "streamoutput":
        if arguments:
            stream_output = arguments[0] == "on"
        else:
            stream_output = not stream_output
        print(f"# {'Now streaming responses' if stream_output else 'No longer streaming responses'}")

    elif command == "web" or command == "websearch":
        if arguments:
            is_web_search = arguments[0] == "on"
        else:
            is_web_search = not is_web_search
        print(f"# {'Web searching activated' if is_web_search else 'We searching deactivated'}")

    elif command == "web-hint" or command == "webhint":
        if arguments:
            web_search_hint = arguments[0] == "on"
        else:
            web_search_hint = not web_search_hint
        print(f"# {'Enabled web hint' if web_search_hint else 'Disabled web hint'}")

    else:
        print("# Unrecognized command")


def stream_response(generator) -> None:
    print("<", end="", flush=True)

    for chunk in generator:
        if chunk is None:
            continue
        print(chunk["token"], end="", flush=True)

    print()


def web_search(generator) -> None:
    print("<", end="", flush=True)

    sources = []
    for chunk in generator:
        if web_search_hint and chunk['type'] == 'webSearch' and chunk['messageType'] == 'update':
            args = chunk['args'][0] if 'args' in chunk else ""
            print(f"🌍 Web Searching | {chunk['message']} {args}")

        elif web_search_hint and chunk['type'] == 'webSearch' and chunk['messageType'] == 'sources' and "sources" in chunk:
            sources = chunk['sources']

        elif chunk['type'] == 'stream':
            print(chunk['token'], end="", flush=True)

    if web_search_hint and len(sources) > 0:
        print("\n# Sources:")
        for i in range(len(sources)):
            print(f"  {i+1}. {sources[i]['title']} - {sources[i]['link']}")

    print()


def get_arguments() -> tuple:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-u",
        type=str,
        help="Your huggingface account's email"
    )
    parser.add_argument(
        "-p",
        action="store_true",
        help="Require Password to login"
    )
    parser.add_argument(
        "-s",
        action="store_true",
        help="Option to enable streaming mode output"
    )
    parser.add_argument(
        "-c",
        action="store_true",
        help="Continue the previous conversation"
    )
    args = parser.parse_args()
    return args.u, args.p, args.s, args.c

# ...


def cli():
    global stream_output, is_web_search, web_search_hint , continued_conv

    print("""
-------HuggingChat-------
Official Site: https://huggingface.co/chat
1. AI is an area of active research with known problems such as biased generation and misinformation.
2. Do not use this application for high-stakes decisions or advice.
Continuing to use means that you accept the above point(s)
    """)

    EMAIL, FORCE_INPUT_PASSWORD, stream_output, continued_conv = get_arguments()

    # Check if the email is in the environment variables or given as an argument
    if EMAIL is None:
        EMAIL = os.getenv("EMAIL")

        if EMAIL is None:
            raise Exception("No email specified. Please use '-u' or set the EMAIL environment variable.")

    # Attempt to load cookies from directory
    try:
        cookies = Login(EMAIL).loadCookiesFromDir()
    except Exception:
        cookies = None

    # If we could not find cookies or the "inputpass" argument is true
    # then ask for the password
    if cookies is None or FORCE_INPUT_PASSWORD:
        if ENVIRONMENT_PASSWORD is not None and not FORCE_INPUT_PASSWORD:
            PASSWORD = ENVIRONMENT_PASSWORD
        else:
            PASSWORD = getpass.getpass("Password: ")

        print("Logging in...")
        sign = Login(EMAIL, PASSWORD)
        cookies = sign.login()
        sign.saveCookiesToDir()

    print(f"Signed in as {EMAIL} .. attempting to login")

    chatbot = ChatBot(cookies=cookies)

    print("Login successfully! 🎉\nYou can input `/help` to open the command menu.\n")

    if continued_conv:
        ids = chatbot.get_remote_conversations(replace_conversation_list=True)

        conversation_dict = dict(enumerate(ids, start=1))

        # delete the conversation that was initially created on chatbot init
        target_id = conversation_dict[1]
        chatbot.delete_conversation(target_id)

        # switch to the most recent remote conversation
        target_id = conversation_dict[2]
        chatbot.change_conversation(target_id)

        print(f"Switched to Previous conversation with ID: {target_id}\n")

    while True:
        userInput = input("> ").strip()

        if len(userInput) == 0:
            continue
        if userInput.lower() in EXIT_COMMANDS:
            break

        # If the input starts with a slash then we know it is a command
        if userInput.startswith("/"):
            try:
                handle_command(chatbot, userInput)
            except Exception as e:
                print("An error occurred while processing your command: " + str(e))
                traceback.print_exc()

            continue

        # either start a web search or a normal response
        if is_web_search:
            res = chatbot.chat(userInput, stream=True, _stream_yield_all=True, web_search=is_web_search)
            web_search(res)

        else:
            if stream_output:
                res = chatbot.chat(userInput)
                stream_response(res)
            else:
                print("< " + chatbot.chat(userInput).wait_until_done().strip())

if __name__ == '__main__':
    cli()



class ModelOverloadedError(Exception):
    """
    HF Model Overloaded Error

    Raised when hf return response `{"error":"Model is overloaded", "error_type":"overloaded"}`
    """
    pass


class ChatBotInitError(Exception):
    """
    ChatBot Init Error

    Raised when chatbot init failed
    """
    pass


class CreateConversationError(Exception):
    """
    Create Conversation Error

    Raised when create conversation failed
    """
    pass


class InvalidConversationIDError(Exception):
    """
    Invalid Conversation ID Error

    Raised when using a invalid conversation id
    """
    pass


class DeleteConversationError(Exception):
    """
    Delete Conversation Error

    Raised when delete conversation failed
    """
    pass


class ChatError(Exception):
    """
    Chat Error

    Raised when chat failed
    """
    pass



from typing import Generator, Union, List

from .types.tool import Tool
from .types.file import File
from .types.message import Conversation
from .exceptions import ChatError, ModelOverloadedError
import json


class ResponseTypes:
    FINAL = "finalAnswer"
    STREAM = "stream"
    TOOL = "tool"
    FILE = "file"
    WEB = "webSearch"
    STATUS = "status"


class MessageStatus:
    PENDING = 0
    RESOLVED = 1
    REJECTED = 2


MSGTYPE_ERROR = "error"


class WebSearchSource:
    title: str
    link: str

    def __str__(self):
        return json.dumps({
            "title": self.title,
            "link": self.link,
        })


class Message(Generator):
    """
    :Args:
        * g: Generator
        * _stream_yield_all: bool = False
        * web_search: bool = False
        - web_search_sources: list[WebSearchSource] = list()
        - text: str = ""
        - web_search_done: bool = not web_search
        - msg_status: int = MessageStatus.PENDING
        - error: Union[Exception, None] = None

    A wrapper of `Generator` that receives and process the response

    :Example:
    .. code-block:: python

        msg = bot.chat(...)

        # stream process
        for res in msg:
            ... # process
        else:
            if msg.done() == MessageStatus.REJECTED:
                raise msg.error

        # or simply use:
        final = msg.wait_until_done()
    """

    _stream_yield_all: bool = False
    _result_text: str = ""

    gen: Generator

    web_search: bool = False
    web_search_sources: List[WebSearchSource] = []
    web_search_done: bool = not web_search

    tools_used: List[Tool] = []
    files_created: List[File] = []

    msg_status: int = MessageStatus.PENDING
    error: Union[Exception, None] = None

    def __init__(
        self,
        g: Generator,
        _stream_yield_all: bool = False,
        web_search: bool = False,
        conversation: Conversation = None
    ) -> None:
        self.gen = g
        self._stream_yield_all = _stream_yield_all
        self.web_search = web_search
        self.conversation = conversation

    @property
    def text(self) -> str:
        self._result_text = self.wait_until_done()
        return self._result_text

    @text.setter
    def text(self, v: str) -> None:
        self._result_text = v

    def _filterResponse(self, obj: dict):
        if not obj.__contains__("type"):
            if obj.__contains__("message"):
                raise ChatError(f"Server returns an error: {obj['message']}")
            else:
                raise ChatError(f"No `type` and `message` returned: {obj}")

    def __next__(self) -> dict:
        if self.msg_status == MessageStatus.RESOLVED:
            raise StopIteration

        elif self.msg_status == MessageStatus.REJECTED:
            if self.error is not None:
                raise self.error
            else:
                raise Exception("Message status is `Rejected` but no error found")

        try:
            data: dict = next(self.gen)
            self._filterResponse(data)
            data_type: str = data["type"]
            message_type: str = ""

            # set _result_text if this is the final iteration of the chat message
            if data_type == ResponseTypes.FINAL:
                self._result_text = data["text"]
                self.msg_status = MessageStatus.RESOLVED

            # Handle web response type
            elif data_type == ResponseTypes.WEB:
                # gracefully pass unparseable webpages
                if message_type != MSGTYPE_ERROR and data.__contains__("sources"):
                    self.web_search_sources.clear()
                    sources = data["sources"]
                    for source in sources:
                        wss = WebSearchSource()
                        wss.title = source["title"]
                        wss.link = source["link"]
                        self.web_search_sources.append(wss)

            # Handle what is done when a tool completes
            elif data_type == ResponseTypes.TOOL:
                if data["subtype"] == "result":
                    tool = Tool(data["uuid"], data["result"])
                    self.tools_used.append(tool)

            # Handle what is done when a file is created
            elif data_type == ResponseTypes.FILE:
                file = File(data["sha"], data["name"], data["mime"], self.conversation)
                self.files_created.append(file)

            # replace null characters with an empty string
            elif data_type == ResponseTypes.STREAM:
                data["token"] = data["token"].replace('\u0000', '')

            elif "messageType" in data:
                message_type: str = data["messageType"]
                if message_type == MSGTYPE_ERROR:
                    self.error = ChatError(data["message"])
                    self.msg_status = MessageStatus.REJECTED

                if data_type == ResponseTypes.STREAM:
                    self.web_search_done = True

                elif data_type == ResponseTypes.STATUS:
                    pass

            else:
                if "Model is overloaded" in str(data):
                    self.error = ModelOverloadedError(
                        "Model is overloaded, please try again later or switch to another model."
                    )
                    self.msg_status = MessageStatus.REJECTED
                elif data.__contains__(MSGTYPE_ERROR):
                    self.error = ChatError(data[MSGTYPE_ERROR])
                    self.msg_status = MessageStatus.REJECTED
                else:
                    self.error = ChatError(f"Unknown json response: {data}")

            # If _stream_yield_all is True, yield all responses from the server.
            if self._stream_yield_all or data_type == ResponseTypes.STREAM:
                return data
            else:
                return self.__next__()
        except StopIteration:
            if self.msg_status == MessageStatus.PENDING:
                self.error = ChatError(
                    "Stream of responses has abruptly ended (final answer has not been received)."
                )
                raise self.error
            pass
        except Exception as e:
            # print("meet error: ", str(e))
            self.error = e
            self.msg_status = MessageStatus.REJECTED
            raise self.error

    def __iter__(self):
        return self

    def throw(
        self,
        __typ,
        __val=None,
        __tb=None,
    ):
        return self.gen.throw(__typ, __val, __tb)

    def send(self, __value):
        return self.gen.send(__value)

    def get_final_text(self) -> str:
        """
        :Return:
            - self.text
        """
        return self.text

    def get_search_sources(self) -> list:
        """
        :Return:
            - self.web_search_sources
        """
        return self.web_search_sources

    def get_tools_used(self) -> list:
        """
        :Return:
            - self.tools_used
        """
        return self.tools_used

    def get_files_created(self) -> list:
        """
        :Return:
            - self.files_created
        """
        return self.files_created

    def search_enabled(self) -> bool:
        """
        :Return:
            - self.web_search
        """
        return self.web_search

    def wait_until_done(self) -> str:
        """
        :Return:
            - self.text if resolved else raise error

        wait until every response is resolved
        """
        while not self.is_done():
            self.__next__()

        if self.is_done() == MessageStatus.RESOLVED:
            return self._result_text

        elif self.error is not None:
            raise self.error

        else:
            raise Exception("Rejected but no error captured!")

    def is_done(self):
        """
        :Return:
            - self.msg_status

        3 status:
        - MessageStatus.PENDING = 0    # running
        - MessageStatus.RESOLVED = 1   # done with no error(maybe?)
        - MessageStatus.REJECTED = 2   # error raised
        """
        return self.msg_status

    def is_done_search(self):
        """
        :Return:
            - self.web_search_done

        web search result will be set to `done` once the token is received
        """
        return self.web_search_done

    def __str__(self):
        return self.wait_until_done()

    def __getitem__(self, key: str) -> str:
        print("_getitem_")
        self.wait_until_done()

        if key == "text":
            return self.text

        elif key == "web_search":
            return self.web_search

        elif key == "web_search_sources":
            return self.web_search_sources

    def __add__(self, other: str) -> str:
        self.wait_until_done()
        return self.text + other

    def __radd__(self, other: str) -> str:
        self.wait_until_done()
        return other + self.text

    def __iadd__(self, other: str) -> str:
        self.wait_until_done()
        self.text += other
        return self.text



from dataclasses import dataclass


class Model:
    def __init__(
        self,
        id: str = None,
        name: str = None,
        displayName: str = None,
        preprompt: str = None,
        promptExamples: list = None,
        websiteUrl: str = None,
        description: str = None,
        datasetName: str = None,
        datasetUrl: str = None,
        modelUrl: str = None,
        parameters: dict = None,
    ):
        """
        Returns a model object
        """

        self.id: str = id
        self.name: str = name
        self.displayName: str = displayName

        self.preprompt: str = preprompt
        self.promptExamples: list = promptExamples
        self.websiteUrl: str = websiteUrl
        self.description: str = description

        self.datasetName: str = datasetName
        self.datasetUrl: str = datasetUrl
        self.modelUrl: str = modelUrl
        self.parameters: dict = parameters

    def __str__(self) -> str:
        return self.id



from .message import Conversation


class File:
    '''
    Class used to represent files created by the model
    '''

    def __init__(self, sha: str, name: str, mime: str, conversation: Conversation):
        self.sha = sha
        self.name = name
        self.mime = mime

        self.conversation = conversation
        self.url = self.get_url()

    def get_url(self) -> str:
        """
        Gets the url for the given file
        """

        return f"https://huggingface.co/chat/conversation/{self.conversation.id}/output/{self.sha}"

    def download_file(self, chatBot) -> bytes:
        """
        Downloads the given file
        """

        r = chatBot.session.get(self.url)
        return r.content

    def __str__(self) -> str:
        return f"File(url={self.url}, sha={self.sha}, name={self.name}, mime={self.mime})"
